@@ -50,8 +50,7 @@ mongoc_bulk_operation_new (bool ordered)
    mongoc_bulk_operation_t *bulk;
 
    bulk = (mongoc_bulk_operation_t *) bson_malloc0 (sizeof *bulk);
-   bulk->flags.bypass_document_validation =
-      MONGOC_BYPASS_DOCUMENT_VALIDATION_DEFAULT;
+   bulk->flags.bypass_document_validation = false;
    bulk->flags.ordered = ordered;
    bulk->server_id = 0;
 
@@ -267,7 +266,7 @@ mongoc_bulk_operation_remove (mongoc_bulk_operation_t *bulk, /* IN */
    BULK_EXIT_IF_PRIOR_ERROR;
 
    if (!mongoc_bulk_operation_remove_many_with_opts (
-         bulk, selector, NULL, error)) {
+          bulk, selector, NULL, error)) {
       MONGOC_WARNING ("%s", error->message);
    }
 
@@ -290,7 +289,7 @@ mongoc_bulk_operation_remove_one (mongoc_bulk_operation_t *bulk, /* IN */
    BULK_EXIT_IF_PRIOR_ERROR;
 
    if (!mongoc_bulk_operation_remove_one_with_opts (
-         bulk, selector, NULL, error)) {
+          bulk, selector, NULL, error)) {
       MONGOC_WARNING ("%s", error->message);
    }
 
@@ -379,12 +378,7 @@ mongoc_bulk_operation_insert_with_opts (mongoc_bulk_operation_t *bulk,
    }
 
    _mongoc_write_command_init_insert (
-      &command,
-      document,
-      opts,
-      bulk->flags,
-      bulk->operation_id,
-      !mongoc_write_concern_is_acknowledged (bulk->write_concern));
+      &command, document, &insert_opts.extra, bulk->flags, bulk->operation_id);
 
    _mongoc_array_append_val (&bulk->commands, command);
 
@@ -409,6 +403,7 @@ _mongoc_bulk_operation_update_append (
    bson_t opts;
    bool has_collation;
    bool has_array_filters;
+   bool has_update_hint;
 
    bson_init (&opts);
    bson_append_bool (&opts, "upsert", 6, update_opts->upsert);
@@ -416,7 +411,7 @@ _mongoc_bulk_operation_update_append (
 
    has_array_filters = !bson_empty0 (array_filters);
    if (has_array_filters) {
-      bson_append_document (&opts, "arrayFilters", 12, array_filters);
+      bson_append_array (&opts, "arrayFilters", 12, array_filters);
    }
 
    has_collation = !bson_empty (&update_opts->collation);
@@ -424,6 +419,11 @@ _mongoc_bulk_operation_update_append (
       bson_append_document (&opts, "collation", 9, &update_opts->collation);
    }
 
+   has_update_hint = !!(update_opts->hint.value_type);
+   if (has_update_hint) {
+      bson_append_value (&opts, "hint", 4, &update_opts->hint);
+   }
+
    if (extra_opts) {
       bson_concat (&opts, extra_opts);
    }
@@ -433,6 +433,7 @@ _mongoc_bulk_operation_update_append (
          &bulk->commands, mongoc_write_command_t, bulk->commands.len - 1);
       if (last->type == MONGOC_WRITE_COMMAND_UPDATE) {
          last->flags.has_collation |= has_collation;
+         last->flags.has_update_hint |= has_update_hint;
          last->flags.has_multi_write |= update_opts->multi;
          _mongoc_write_command_update_append (last, selector, document, &opts);
          bson_destroy (&opts);
@@ -445,6 +446,7 @@ _mongoc_bulk_operation_update_append (
 
    command.flags.has_array_filters = has_array_filters;
    command.flags.has_collation = has_collation;
+   command.flags.has_update_hint = has_update_hint;
    command.flags.has_multi_write = update_opts->multi;
 
    _mongoc_array_append_val (&bulk->commands, command);
@@ -574,8 +576,10 @@ mongoc_bulk_operation_update (mongoc_bulk_operation_t *bulk,
       BSON_APPEND_BOOL (&opts, "upsert", upsert);
    }
 
-   mongoc_bulk_operation_update_many_with_opts (
-      bulk, selector, document, &opts, error);
+   if (!mongoc_bulk_operation_update_many_with_opts (
+          bulk, selector, document, &opts, error)) {
+      MONGOC_WARNING ("%s", error->message);
+   }
 
    bson_destroy (&opts);
 
@@ -603,7 +607,7 @@ mongoc_bulk_operation_update_one (mongoc_bulk_operation_t *bulk,
    BSON_APPEND_BOOL (&opts, "upsert", upsert);
 
    if (!mongoc_bulk_operation_update_one_with_opts (
-         bulk, selector, document, &opts, error)) {
+          bulk, selector, document, &opts, error)) {
       MONGOC_WARNING ("%s", error->message);
    }
 
@@ -787,6 +791,11 @@ mongoc_bulk_operation_execute (mongoc_bulk_operation_t *bulk, /* IN */
                                      &bulk->result);
 
       bulk->server_id = server_stream->sd->id;
+      /* If a retryable error occurred and a new primary was selected, use it in
+       * subsequent commands. */
+      if (bulk->result.retry_server_id) {
+         bulk->server_id = bulk->result.retry_server_id;
+      }
 
       if (bulk->result.failed &&
           (bulk->flags.ordered || bulk->result.must_stop)) {
@@ -930,7 +939,5 @@ mongoc_bulk_operation_set_bypass_document_validation (
 {
    BSON_ASSERT (bulk);
 
-   bulk->flags.bypass_document_validation =
-      bypass ? MONGOC_BYPASS_DOCUMENT_VALIDATION_TRUE
-             : MONGOC_BYPASS_DOCUMENT_VALIDATION_FALSE;
+   bulk->flags.bypass_document_validation = bypass;
 }
