@@ -14,17 +14,14 @@
  * limitations under the License.
  */
 
-
-#if !defined(MONGOC_COMPILATION)
-#error "Only <mongoc.h> can be included directly."
-#endif
+#include "mongoc-prelude.h"
 
 
 #ifndef MONGOC_SCRAM_PRIVATE_H
 #define MONGOC_SCRAM_PRIVATE_H
 
 
-#include <bson.h>
+#include <bson/bson.h>
 #include "mongoc-crypto-private.h"
 
 
